
/*
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *  MAIN PROGRAM FOR MG BENCHMARK - HPSPMD VERSION
 *
 *  SUPPLIED BY   MAX LEMKE
 *                SUPRENUM GMBH
 *                HOHE STRASSE 73
 *                D-5300 BONN 1
 *
 *  VERSION FEBRUARY 15, 1989
 *
 *  REFERENCES:
 *            - K. STUEBEN, U. TROTTENBERG: MULTIGRID METHODS:
 *              FUNDAMENTAL ALGORITHMS, MODEL PROBLEM ANALYSIS
 *              AND APPLICATIONS, DECEMBER 84, GMD-STUDIEN NR. 96,
 *              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG MBH,
 *              BONN
 *
 *            - M. LEMKE: EXPERIMENTS WITH A VECTORIZED MULTIGRID
 *              POISSON SOLVER ON THE CDC CYBER 205, CRAY X-MP AND
 *              FUJITSU VP 200;
 *              NOVEMBER 85, ARBEITSPAPIERE DER GMD, NR. 179,
 *              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG MBH,
 *              BONN
 *
 *            - M. LEMKE: ERFAHRUNGEN MIT MEHRGITTERVERFAHREN FUER
 *              HELMHOLTZ-AEHNLICHE PROBLEME AUF VEKTORRECHNERN
 *              UND MEHRPROZESSOR-VEKTORRECHNERN
 *              DECEMBER 87, ARBEITSPAPIERE DER GMD, NR. 278,
 *              GESELLSCHAFT FUER MATHEMATIK UND DATENVERARBEITUNG MBH,
 *              BONN
 *
 *  AUTHORS: FIRST VERSION:
 *              K. WITSCH, H. FOERSTER (1984, GMD)
 *           VECTORIZED (AND MULTITASKING) VERSION:
 *              M. LEMKE (1984 - 1987, GMD)
 *           GENESIS BENCHMARK VERSION:
 *              M. LEMKE (1989, SUPRENUM GMBH)
 *           SUBSET FORTRAN 90 REWRITE:
 *              D.B. CARPENTER (1993, UNIV. OF SOUTHAMPTON)
 *           HPJAVA REWRITE:
 *              D.B. CARPENTER (2000, UNIV. OF SYRACUSE)
 *
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *  THE TIMER (DOUBLE PRECISION IN SECONDS) IS CALLED IN MG00D.
 *  THE MACHINE DEPENDENT TIMING ROUTINE HAS TO BE INCLUDED.
 *
 *  ATTENTION: ONLY THE MG KERNEL IS VECTORIZED. SOME OTHER ROUTINES
 *             ARE PURELY SEQUENTIAL. THEY NEED NOT BE VECTORIZED
 *             BECAUSE THEY ARE NOT INCLUDED IN TIMING.
 *
 *             THE FMG OPTION IS NOT AVAILABLE IN THIS VERSION
 *
 *  FOR NUMERICAL VERIFICATION:
 *     THE AVERAGE CONVERGENCE RATE SHOLD BE SOMEWHERE IN BETWEEN
 *     0.02 AND 0.07.
 *
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 */

class PDE2 {

    final int MMAX = 8;
    final int NITER = 60;

    public static void main(String args []) {

        // Definition of finest gridsize: (2**MMAX+1)*(2**MMAX+1)

        // Results are obtained for the finest grid ranging from 4 to MMAX.
        // MMAX should be chosen as large as possible, depending on memory
        // capacities of the machine.

        sec2 = 1.0;  // does nothing?

        Adlib.gprintln(
            "************* MULTIGRID BENCHMARK ******************* ");

        p = new Procs2();

	on(p) {

	    // Loop over the finest grids in order to get results for
            // different problem sizes

	    int ncycle = NITER;

	    for(int m = 3; m <= MMAX; m++) {

		try {
                    int ny1 = 2, ny2 = 1, nfmg = 0, igam = 1;
		    mg00d (m, ny1, ny2, nfmg, ncycle, igam,
                           new F(), new G());

		} catch(MGException e) {

		    e.printStackTrace();
		    HPspmd.abort ();
		}

		int n  = 1 << m;
		int np = n + 1;

		double dif = getdif (m, G);
		double def = getdef (m)

		double convr = Math.pow (def/def1, 0.1);

		Adlib.gprintln ("NUMBER OF POINTS PER DIRECTION = " + np);
		Adlib.gprintln ("GRIDSIZE OF SQUARE GRID =        " +
                                                                 np * np);
		Adlib.gprintln ("MAXIMUM NORM OF THE ERROR  =     " + dif);
		Adlib.gprintln ("MAXIMUM NORM OF THE DEFECT =     " + def);
		Adlib.gprintln ("AVERAGE CONV. RATE             = " + convr);

                int nop = 1;

		for(int lev = 2; lev <= m; lev++)
		    nop +=
		        (int) (((1 << lev) - 1) * ((1 << lev) - 1) *
                               (2.75 + 9.0)) +
		        ((1 << (lev - 1)) - 1) * ((1 << (lev - 1)) - 1) * 7
		        ((1 << (lev - 1)) - 1) * ((1 << (lev - 1)) - 1) * 6;
	     
		nop += (int) (((1 << M) - 1) * ((1 << M) - 1) * 1.75);

                double rmflop = ((nop / 1000000.0) / sec2) * ncycle

		Adlib.gprintln("TIME FOR " + ncycle +
                               " CYCLES (SEC.):" + sec2);
		Adlib.gprintln("TIME FOR     1 CYCLE (SEC.):" + sec2/ncycle);
		Adlib.gprintln("NUMBER OF OPERATIONS:       " + nop);
		Adlib.gprintln("PERFORMANCE IN MFLOP/S:     " + rmflop);

		Adlib.gprintln(
		  "+++++++++++++++++++++++++++++++++++++++++++++++++++++ ");

		ncycle = Math.max (ncycle / 4, 10);
	    }
	}
    }

 
    /*
     *  MULTIGRID MODULE FOR THE FAST SOLUTION OF POISSON'S EQUATION
     *  WITH DIRICHLET BOUNDARY CONDITIONS ON THE UNIT SQUARE
     *
     *       DIFFERENTIAL EQUATION :   -DELTA U(X,Y) = F(X,Y)
     *       BOUNDARY CONDITION :             U(X,Y) = G(X,Y)
     *
     *  CYCLE STRUCTURE:
     *
     *             COARSENING: H, 2H, 4H, ...
     *             DIFFERENCE OPERATORS: USUAL 5-POINT STARS ON ALL GRIDS
     *             RELAXATION: RED-BLACK
     *             FINE-TO-COARSE: HALF INJECTION
     *             COARSE-TO-FINE: BILINEAR INTERPOLATION
     *
     *  FULL MULTIGRID MODE:
     *
     *             FULL MULTIGRID INTERPOLATION USES GRID EQUATION
     *             (4-TH ORDER)
     *
     *  INPUT:
     *
     *     M       NUMBER OF GRIDS (0 < M < 11). FOR GIVEN M:
     *             NUMBER OF P0INTS ON THE FINEST GRID (INCLUDING BOUNDARY
     *             POINTS) = 2**M + 1 IN BOTH DIRECTIONS
     *     NY1     NUMBER OF RELAXATIONS BEFORE COARSE-GRID CORRECTION
     *             (NY1 > 0)
     *     NY2     NUMBER OF RELAXATIONS AFTER COARSE-GRID CORRECTION
     *             (NY2 > 0)
     *     NCYCLE  NUMBER OF MULTIGRID ITERATIONS (NCYCLE > 0)
     *     NFMG    .EQ. 0 : NCYCLE MULTIGRID ITERATIONS ARE PERFORMED
     *             .NE. 0 : FMG-VERSION IS PERFORMED PLUS NCYCLE-1
     *                      ADDITIONAL MG ITERATIONS AFTERWARDS
     *     IGAM    TYPE OF CYCLING (IGAM > 0).
     *             E.G.: IGAM=1 FOR V-CYCLES, IGAM=2 FOR W-CYCLES
     *     W       DOUBLE PRECISION WORK ARRAY OF DIMENSION IDIM
     *     IDIM    DIMENSION OF W. APPROXIMATELY: IDIM > 2.8*4**M
     *
     *  EXTERNALS:
     *
     *     F       DOUBLE PRECISION FUNCTION F(X,Y), RIGHT HAND SIDE OF
     *             THE DIFFERENTIAL EQUATION
     *     G       DOUBLE PRECISION FUNCTION G(X,Y), BOUNDARY VALUES
     *
     *  REMARK:    GRID #1 AND #M ARE THE FINEST AND COARSEST GRID USED,
     *             RESPECTIVELY
     *
     *  OUTPUT:
     *
     *     IER     ERROR INDICATOR
     *             = 0  NO ERRORS
     *             = 1  INSUFFICIENT MEMORY , I.E. IDIM TOO SMALL.
     *                  IN THIS CASE IDIM IS USED AS OUTPUT PARAMETER
     *                  TO SHOW THE MINIMAL DIMENSION
     *             = 2  M, NY1, NY2, NCYCLE OR IGAM WRONG
     *     JDIM    ONLY IN CASE IER=1 : MINIMAL LENGTH OF W
     *     INITF   CF. DESCRIPTION OF W
     *     W       W CONTAINS THE DISCRETE APPROXIMATION TO THE GIVEN
     *             BOUNDARY VALUE PROBLEM ON THE FINEST GRID. THE GRID
     *             VALUES ARE STORED ROWWISE FROM LEFT TO RIGHT AND
     *             FROM BOTTOM TO TOP.
     *             I.E. THE GRID VALUE CORRESPONDING TO THE GRID POINT
     *             (XI,YJ) :
     *
     *                  XI = (I-1)*H,  YJ = (J-1)*H   (H=1/N, N=2**M)
     *
     *             IS STORED AT
     *
     *                     W((J-1)*(N+1)+I)     (0 < I,J < N+2)
     *
     *             THE CORRESPONDING VALUES OF THE RIGHT HAND SIDE ARE
     *             STORED IN THE SAME MANNER AT
     *
     *                     W((J-1)*(N+1)+I+INITF).
     *
     *             THE REMAINING STORAGE CONTAINS COARSE GRID VALUES.
     *
     *  REMARK:    THIS IS A SPECIALIZED PROGRAM FROM THE MG00 PROGRAM
     *             COLLECTION. SEE: FOERSTER, WITSCH IN
     *             'ON EFFICIENT MULTIGRID SOFTWARE FOR ELLIPTIC
     *             PROBLEMS ON RECTANGULAR DOMAINS', MATH. COMPUT.
     *             SIMULATION XXIII, PP. 293-298, 1981
     */

    static void mg00d (int m, int ny1, int ny2, int nfmg,
                       int ncycle, int igam,
                       Function f, Function g) throws MGException {

        // Check of error conditions (wrong parameters)

        if((m <= 0) || (m >= 110) || (ny1 <= 0) || (ny2 <= 0) ||
                (ncycle <= 0) || (igam <= 0)) {
            throw new MGException("*** ERROR, IER=" + 2 + 
                                  " IDIM=" + 0 + " ***");
        }

	// Initialise stack...

	inistk (m);

	// lin = level from which initial guess propagated down

        int lin = (nfmg == 0 ? 1 : m);

	// Set up all grid values needed

	init (m, lin, f, g);

	// (Full) multigrid procedure...

	// Initial guess from coarsest grid

	for(int lev = lin; lev >= 2; lev--) {
	    mgi (lev, m, ny1, ny2, igam);

	    fmgdn (nps [m - lev + 1], us [m - lev + 1], fs [m - lev + 1],
		   nps [m - lev], us [m - lev]);
	}

	// Multigrid cycles

	def1 = getdef (m);

	//----> TIMING

        double sec1 = timer();

	for(int it = 1; it <= ncycle; i++)
	    mgi (1, m, ny1, ny2, igam);

	sec2 = timer() - sec1;

	//----> TIMING
    }


    /*
     * Initialise stack
     */

    static void inistk (int m) {

        nps = new int [m];

        us  = new double [m] [[,]];
        fs  = new double [m] [[,]];

        for(int l = 0; l < m; l++) {
            final int np = (1 << (l + 1)) + 1;

            Range x = new ExtendedBlockRange(np, p.dim(0), 1);
            Range y = new ExtendedBlockRange(np, p.dim(1), 1);

            nps [l] = np;

            us [l] = new double [[x, y]];
            fs [l] = new double [[x, y]];
        }
    }


    /*
     * Computes initial values on all grids.
     * lin = level from which initial guess propagated down
     */

    static void init (int m, int lin, Function f, Function g) {

        double h = 1.0 / (1 << m);

        // Initialisation on finest level

        init1 (nps [m - 1], h, f, g, us [m - 1], fs [m - 1]);
       
        // Transfer of `f' and `u' to coarser grids

        for(int l = 2; l <= lin; l++)
            trans (nps [m - l], nps [m - l + 1], us [m - l], fs [m - 1],
                   us [m - l + 1], fs [m - l + 1]);

        // Put zero to boundary values for multigrid corrections

        for(int l = lin + 1; l <= m; l++) {
            putzb (nps [m - l], us [m - l]);
        }
    }


    /*
     *  One multigrid iteration step
     */

    static void mgi(int lev, int m, int ny1, int ny2, int igam) {

        int k;
        int [] icgam = new int [10], igaml = new int [10];
        boolean iz;

        for(k = lev; k <= m; k++) {
          if((k == lev) || (k == m)) {
            igaml [k - 1] = 1;
          } else {
            igaml [k - 1] = igam;
          }

          icgam [k - 1] = 0;
        }

        k = lev;

        while (icgam [k - 1] < igaml [k - 1]) {

            // "Up"

            while (k < m) {
                iz = k > lev && icgam [k - 1] == 0;

                mgiup (nps [m - k - 1], fs [m - k - 1],
                       nps [m - k], us [m - k], fs [m - k], ny1, iz);

                icgam [k - 1]++;
                k++;
            }

            // Exact solution on coarsest grid (k = m)

            relax (1, (m > lev && icgam [m - 1] == 0),
                   nps [0], us [0], fs [0]);
            
            icgam [m - 1]++;

            // "Down"

            while ((k > lev) && (icgam [k - 1] == igaml [k - 1])) {
                icgam [k - 1] = 0;
                k--;

                mgidn (nps [m - k], us [m - k], fs [m - k],
                       nps [m - k - 1], us [m - k - 1], ny2);
            }
        }
    }


    /*
     *  Extract maximum norm of the defect from stack
     */

    static double getdef (int m) {
        return defmx (us [m - 1], fs [m - 1], nps [m - 1]);
    }


    /*
     *  Extract maximum norm of the difference from stack
     */

    static double getdif (int m, Function sol) {
        return difmx (us [m - 1], nps [m - 1], sol);
    }


    /*
     *  One "down" step in full multigrid
     */

    static void fmgdn (int npf, double uf [[,]], double frf [[,]],
                       int npc, double uc [[,]]) {

        int4 (npc, npf, uc, uf, frf);
        putzb (npc, uc);
    }


    /*
     *  One "up" step in multigrid iteration.
     */

    static void mgiup (int npc, double frc [[,]], int npf,
                       double uf [[,]], double frf [[,]], int ny1,
                       boolean iz) {

        // Relaxations before coarse-grid correction...

        relax (ny1, iz, npf, uf, frf);

        // Residual transfer to next coarser grid...

        restr (npc, npf, frc, uf, frf);
    }

    /*
     *  One "down" step in multigrid iteration.
     */

    static void mgidn (int npf, double uf [[,]], double frf [[,]],
                       int npc, double uc [[,]], int ny2) {

        // Linear interpolation to next finer grid

        int2a (npc, npf, uc, uf);

        // Relaxation after CGC

        relax (ny2, false, npf, uf, frf);
    }


    /*
     *  Computes initial values on the finest grid
     */

    static void init1 (int np, double h, Function f, Function g,
                       double u [[,]], double fr [[,]]) {

        Range xr = fr.rng(0), yr = fr.rng(1);

        double h2 = h * h;
        int n = np - 1;

        overall (j = yr for 1 : n - 1) {
            double y = h * j`;

            at (i = xr [0]) u [i, j] = g.value (0.0, y);

            overall (i = xr for 1 : n - 1) {
                double x = h * i`;

                u  [i, j] = 0.0;
                fr [i, j] = h2 * f.value (x, y);
            }

            at (i = xr [n]) u [i, j] = g.value (1.0, y);
        }

        overall (i = xr for :) {
            double x = h * i`;

            at (j = yr [0]) u [i, j] = g.value (x, 0.0);
            at (j = yr [n]) u [i, j] = g.value (x, 1.0);
        }
    }


    /*
     *  Coarse-to-fine transfer (bilinear interpolation) and correction.
     *  (This modifies odd sites only of `uf')
     */

    static void int2a (int npc, int npf, double uc [[,]], double uf [[,]]) {

        Range xf = uf.rng(0), yf = uf.rng(1);

        double [[,]] tf = new double [[xf, yf]];

        int nf = npf - 1;

        Adlib.remap(tf [[0 : nf : 2, 0 : nf : 2]], uc)

        Adlib.writeHalo(tf);

        overall(i = xf for 1 : nf - 1 : 2)
            overall(j = yf for 2 : nf - 2 : 2)
                uf [i, j] += 0.5 * (tf [i - 1, j] + tf [i + 1, j];

        overall(i = xf for 2 : nf - 2 : 2)
            overall(j = yf for 1 : nf - 1 : 2)
                uf [i, j] += 0.5 * (tf [i, j - 1] + tf [i, j + 1]);
    }


    /*
     *  4-th order FMG-interpolation
     */

    static void int4 (int npc, int npf,
                      double uc [[,]], double uf [[,]], double frf [[,]]) {

        Range xf = uf.rng(0), yf = uf.rng(1);

        double [[,]] tf = new double [[xf, yf]];

        int nf = npf - 1;

        // Transfer of coarse grid values to the fine grid

        Adlib.remap(uf [[0 : nf : 2, 0 : nf : 2]], uc)

        // Computation of the remaining even points by using the
        // rotated 5-point star

        Adlib.writeHalo(uf);  // (Need only write in y direction.)

        overall(i = xf for 0 : nf : 2)
            overall(j = yf for 1 : nf - 1 : 2)
                tf [i, j] = uf [i, j - 1] + uf [i, j + 1];

        Adlib.writeHalo(tf);  // (Need only write in x direction.)

        overall(i = xf for 1 : nf - 1 : 2)
            overall(j = yf for 1 : nf - 1 : 2)
                uf [i, j] = 0.25 * (2.0 * frf [i, j] +
                                    tf [i - 1, j] + tf [i + 1, j]);

        // Computation of the odd points by one half (odd) relaxation step
        // **** WHAT?? -VE ITMAX DOESN'T DO THAT !! ****

        relax (-1, false, npf, uf, frf);
    }

    /*
     *  Puts zero to boundary of coarser grids
     */

    static void putzb (int npc, double uc [[,]]) {

        int nc = npc - 1;

        HPspmd.init(uc [[ 0, 1 : nc - 1]], 0.0);
        HPspmd.init(uc [[nc, 1 : nc - 1]], 0.0);
        HPspmd.init(uc [[:,  0]], 0.0);
        HPspmd.init(uc [[:, nc]], 0.0);
    }

    /*
     *  Red black relaxation on grid npf
     */

    static void relax (int itmax, boolean iz, int npf,
                       double uf [[,]], double ff [[,]]) {

        Range xf = ff.rng(0), yf = ff.rng(1);

        int nxf = nxp - 1;
        int nyf = nyp - 1;

        for(int it = 1; it <= itmax; it++) {

	    if (iz && (it <= 1)) {

		// Zero starting values
		// Relaxation of red points

		overall (i = xf for 1 : nxf - 1 : 2)
		    overall (j = yf for 1 : nyf - 1 : 2)
			uf [i, j] = ff [i, j] * 0.25;

		overall (i = xf for 2 : nxf - 1 : 2)
		    overall (j = yf for 2 : nyf - 1 : 2)
			uf [i, j] = ff [i, j] * 0.25;
	    } else {

		// Non-zero starting values
		// Relaxation of red points

                Adlib.writeHalo(uf);

		overall (i = xf for 1 : nxf - 1 : 2)
		    overall (j = yf for 1 : nyf - 1 : 2)
			uf [i, j] = 0.25 * (ff [i, j] +
					    uf [i - 1, j] + uf [i + 1, j] +
					    uf [i, j - 1] + uf [i, j + 1]);

		overall (i = xf for 2 : nxf - 2 : 2)
		    overall (j = yf for 2 : nyf - 2 : 2)
			uf [i, j] = 0.25 * (ff [i, j] +
					    uf [i - 1, j] + uf [i + 1, j] +
					    uf [i, j - 1] + uf [i, j + 1]);
	    }

	    // Relaxation of black points

            Adlib.writeHalo(uf);

	    overall (i = xf for 2 : nxf - 2 : 2)
		overall (j = yf for 1 : nyf - 1 : 2)
		    uf [i, j] = 0.25 * (ff [i, j] +
					uf [i - 1, j] + uf [i + 1, j] +
					uf [i, j - 1] + uf [i, j + 1]);

	    overall (i = xf for 1 : nxf - 1 : 2)
		overall (j = yf for 2 : nyf - 2 : 2)
		    uf [i, j] = 0.25 * (ff [i, j] +
					uf [i - 1, j] + uf [i + 1, j] +
					uf [i, j - 1] + uf [i, j + 1]);
        }
    }


    /*
     *  Computation of the defect and fine-to-coarse transfer
     *  (half-injection)
     */

    static void restr (int npc, int npf,
                       double fc [[,]], double uf [[,]], double ff [[,]]) {

        Range xf = ff.rng(0), ff = ff.rng(1);

        double [[,]] tf = new double [[xf, yf]];

        int nc = npc - 1, nf = npf - 1;

        Adlib.writeHalo(uf);

        overall(i = xf for 2 : nf - 2 : 2)
            overall(j = yf for 2 : nf - 2 : 2)
                tf [i, j] += 2.0 * (ff [i, j] - 4.0 * uf [i, j] +
                                    uf [i - 1, j] + uf [i + 1, j] +
                                    uf [i, j - 1] + uf [i, j + 1]);

        Adlib.remap(fc [[1 : nc - 1, 1 : nc - 1]],
                    tf [[2 : nf - 2 : 2, 2 : nf - 2 : 2]]);
    }


    /*
     *  Transfer of `f' and `u' from grid `npf' to `npc'
     */

    static void trans (int npc, int npf,
                       double uc [[,]], double fc [[,]],
                       double uf [[,]], double ff [[,]]) {

        Range xc = fc.rng(0), fc = fc.rng(1);

        int nc = npc - 1, nf = npf - 1;

        Adlib.remap(uc, uf [[0 : nf : 2, 0 : nf : 2]]);

        Adlib.remap(fc [[1 : nc - 1, 1 : nc - 1]],
                    ff [[2 : nf - 2 : 2, 2 : nf - c : 2]]);

        overall(i = xc for 1 : nc - 1)
            overall(j = yc for 1 : nc - 1)
                fc [i, j] *= 4.0;
    }

    /*
     *  Computes the maximum norm of the defect
     */

    static double defmx (double uc [[,]], double fc [[,]], int np) {

        Range xc = fc.rng(0), fc = fc.rng(1);

        double [[,]] tc = new double [[xc, yc]];

        int n = np - 1;
        double hsqr = (double) (n * n);

        Adlib.writeHalo(uc);

        overall(i = xc for 1 : n - 1)
            overall(j = yc for 1 : n - 1)
                tc [i, j] = Math.abs((fc [i, j] - 4.0 * uc [i, j] +
				      uc [i, j - 1] + uc [i - 1, j] +
				      uc [i + 1, j] + uc [i, j + 1]) * hsqr);
                   
        return Adlib.maxval(tc);
    }

    /*
     *  Computes the maximum norm of the difference between
     *  `sol' (=solution of the BVP) and the values in `uc'.
     */

    static double difmx (double uc [[,]], int np, Function sol) {

        Range xc = uc.rng(0), yc = uc.rng(1);

        double h = 1.0 / (np - 1);

        overall (j = yc for :) {
            double y = h * j` ;

            overall (i = xc for :) {
                double x = h * i`;

                tc [i, j] = Math.abs(sol.value (x, y) - uc [i, j]);
            }
        }

        return Adlib.maxval(tc);
    }


    /*
     *  Elapsed time in seconds since its first call.
     */

    static Calendar cal;
    static long firstCallTime;

    static double timer() {
        if(cal == null) {
            cal = Calendar.getInstance();
            firstCallTime = cal.getTime().getTime()
        }

        return (double) (cal.getTime().getTime() - firstCallTime) / 1000.0;
    }


    Procs2 p ;

    static int nps [];

    static double [] [[,]] us, fs;

    static double sec2;
    static double def1;

}


interface Function {
    double value(double x, double y) ;
}

class F implements Function {
    double value(double x, double y) {
        return 10.0 * Math.sin (3.0 * x + y);
    }
}

class G implements Function {
    double value(double x, double y) {
        return Math.sin (3.0 * x + y);
    }
}

class MGException extends Exception {
    public MGException() {super() ;}
    public MGException(String message) {super(message) ;}
}


